import matplotlib.pyplot as plt
import numpy as np


def draw_segmentation(csv_data, segment_data, sensor_idx=range(0,3), figure_size=(20, 10)):
    plt.figure(figsize=figure_size)
    for i in range(5):
        plt.subplot(5, 1, i + 1)
        plt.plot(csv_data["x"], csv_data["sensor_data"][:, sensor_idx, i])

    text_up = True
    text_y_basis = 0
    for i in range(segment_data["segment_x"].shape[0]):
        for j in range(5):
            plt.subplot(5, 1, j + 1)
            plt.plot(segment_data["segment_x"][i], segment_data["segment_sensor_data"][i][:, sensor_idx, j])

            sensor_max = np.nanmax(csv_data["sensor_data"][:, sensor_idx, j])
            sensor_min = np.nanmin(csv_data["sensor_data"][:, sensor_idx, j])

            text_x = segment_data["segment_x"][i][0]
            if text_up:
                text_up = False
                text_y = sensor_max - text_y_basis
            else:
                text_up = True
                text_y = sensor_min + text_y_basis

                if text_y_basis <= 0:
                    text_y_basis = 5
                else:
                    text_y_basis = 0

            plt.text(text_x, text_y, segment_data["segment_name"][i])

    plt.show()



import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import io
from tensorboard import program
import datetime
import time


class ConfuseCallback(tf.keras.callbacks.Callback):
    def __init__(self, test_data, test_labels, classes, file_writer, figure_size=(12, 10)):
        super(ConfuseCallback, self).__init__()
        self.test_data = test_data
        self.test_labels = test_labels
        self.classes = classes
        self.file_writer = file_writer
        self.figure_size = figure_size

    def get_precision_recall_plot(self, con_mat):
        precisions = np.array([0] * len(self.classes)).astype('float32')
        recalls = np.array([0] * len(self.classes)).astype('float32')

        for i in range(con_mat.shape[0]):
            tp = con_mat[i, i]
            fn = (con_mat[i, :].sum() - tp)

            fp = (con_mat[:, i].sum() - tp)
            tn = (con_mat.diagonal().sum() - tp)

            # tpr = tp / np.sum(self.test_labels[()] == i)
            # fnr = fn / np.sum(self.test_labels[()] == i)
            # fpr = fp / np.sum(self.test_labels[()] != i)
            # tnr = tn / np.sum(self.test_labels[()] != i)

            precisions[i] = max(0, tp / (tp + fp))
            recalls[i] = max(0, tp / (tp + fn))

        df = pd.DataFrame((self.classes, precisions, recalls)).T
        df.columns = ["Class", "Precision", "Recall"]
        df = pd.melt(df, id_vars="Class", var_name="Type", value_name="Value")

        figure = plt.figure(figsize=self.figure_size)
        sns.barplot(y='Class', x='Value', hue='Type', data=df)
        plt.tight_layout()

        buf = io.BytesIO()
        plt.savefig(buf, format='png')
        plt.close(figure)

        image = tf.image.decode_png(buf.getvalue(), channels=4)
        image = tf.expand_dims(image, 0)

        return image, precisions, recalls

    def on_epoch_end(self, epoch, logs=None):
        try:
            test_pred = self.model.predict(self.test_data)
            test_pred = tf.argmax(test_pred, axis=1)
            accuracy = np.sum(test_pred == self.test_labels) / self.test_labels.shape[0]
            con_mat = tf.math.confusion_matrix(labels=self.test_labels, predictions=test_pred).numpy()
            con_mat_norm = np.around(con_mat.astype('float') / con_mat.sum(axis=1)[:, np.newaxis], decimals=2)

            con_mat_df = pd.DataFrame(con_mat_norm,
                                      index=self.classes,
                                      columns=self.classes)

            precision_recall_image, precisions, recalls = self.get_precision_recall_plot(con_mat)

            figure = plt.figure(figsize=self.figure_size)
            sns.heatmap(con_mat_df, annot=True, cmap=plt.cm.Blues)
            plt.ylabel('True label')
            plt.xlabel('Predicted label')
            plt.title("Accuracy : {:.2f}%, Precision : {:.2f}%, Recall : {:.2f}%".format(accuracy*100, precisions.mean()*100, recalls.mean()*100))

            plt.tight_layout()

            buf = io.BytesIO()
            plt.savefig(buf, format='png')

            plt.close(figure)
            buf.seek(0)
            image = tf.image.decode_png(buf.getvalue(), channels=4)

            image = tf.expand_dims(image, 0)

            # Log the confusion matrix as an image summary.
            with self.file_writer.as_default():
                tf.summary.image("Confusion Matrix", image, step=epoch)
                tf.summary.image("Precision and Recall", precision_recall_image, step=epoch)
        except Exception as e:
            print(e)


class ModelSaverCallback(tf.keras.callbacks.Callback):
    def __init__(self, best_loss=float('inf'), save_root="./", enable=True, epoch=0):
        super(ModelSaverCallback, self).__init__()
        self.best_loss = best_loss
        self.epoch = epoch
        self.save_root = save_root
        self.enable = enable

    def on_epoch_end(self, epoch, logs=None):
        try:
            epoch += self.epoch
            if logs['val_loss'] < self.best_loss:
                # TODO Delete previous saved model if exists
                file_name = '{}my_model_weight_{:04d}_{:03.2f}.h5'.format(self.save_root, epoch, logs['val_loss'])
                print("\nBest loss! saving the model to {} ...".format(file_name))
                self.best_loss = logs['val_loss']

                if self.enable:
                    self.model.save(file_name)
        except Exception as e:
            print(e)


def run_tensorboard(path, port=6006):
    tb = program.TensorBoard()
    tb.configure(argv=[None, '--logdir', path, '--host', '0.0.0.0', '--port', f"{port:}"])
    url = tb.launch()

    print("Running tensorboard on {}".format(url))

    return url


def wait_ctrl_c(pre_msg="Press Ctrl+c to quit Tensorboard", post_msg="\nExit."):
    print(pre_msg)
    try:
        while True:
            time.sleep(3600)
    except KeyboardInterrupt:
        print(post_msg)


def get_tf_callbacks(root,
                     tboard_callback=True, tboard_update_freq='epoch', tboard_histogram_freq=1, tboard_profile_batch=0,
                     confuse_callback=True, label_info=None, test_generator_=None,
                     modelsaver_callback=True, best_loss=float('inf'), save_root=None, best_epoch=0):
    postfix = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_root_ = "{}{}/".format(root, postfix)

    callbacks_ = []

    if tboard_callback:
        callbacks_.append(tf.keras.callbacks.TensorBoard(log_dir="{}fit".format(log_root_),
                                                         histogram_freq=tboard_histogram_freq,
                                                         update_freq=tboard_update_freq,
                                                         profile_batch=tboard_profile_batch)
                         )

    #TODO: generate label_info if not given
    if confuse_callback:
        file_writer = tf.summary.create_file_writer("{}/cm".format(log_root_, postfix))
        callbacks_.append(ConfuseCallback(test_generator_.data['test_data'],
                                          test_generator_.data['test_label'],
                                          list(label_info.keys()),
                                          file_writer)
                          )

    if modelsaver_callback:
        if not save_root:
            save_root = log_root_
        callbacks_.append(
            ModelSaverCallback(best_loss=best_loss, save_root=save_root, epoch=best_epoch)
        )

    return callbacks_, log_root_



import numpy as np


def normalize_data(data, norm_max=1):
    min_v = np.ones(data[0].shape[1:]) * np.inf
    max_v = np.ones(data[0].shape[1:]) * -np.inf

    for d in data:
        for i in range(d.shape[1]):
            for j in range(d.shape[2]):
                min_v[i, j] = min(min_v[i, j], d[:, i, j].min())
                max_v[i, j] = max(max_v[i, j], d[:, i, j].max())

    for d in data:
        for i in range(d.shape[1]):
            for j in range(d.shape[2]):
                d[:, i, j] = ((d[:, i, j] - min_v[i, j]) / (max_v[i, j]-min_v[i, j])) * norm_max

    return data


def reshape_data(data, rotate=True):
    data = data.reshape(data.shape[0], data.shape[1], data.shape[2]*data.shape[3])
    if rotate:
        data = np.rot90(data, axes=(1,2))

    return data


def generate_sliding_window_data(data, label, window_size=300, stride=1):
    n_data = 0
    for d, l in zip(data, label):
        if d.shape[0] < window_size:
            continue

        for i in range(0, d.shape[0] - window_size, stride):
            n_data += 1

    w_data = np.zeros((n_data, window_size, data[0].shape[1], data[0].shape[2]))
    w_label = np.zeros((n_data, ))
    idx = 0
    for d, l in zip(data, label):
        if d.shape[0] < window_size:
            continue

        for i in range(0, d.shape[0]-window_size, stride):
            w_data[idx] = d[i:(i+window_size)]
            w_label[idx] = l
            idx += 1

    return w_data, w_label



try:
    from ..utils.data_loader import read_mat_file
except:
    from utils.data_loader import read_mat_file

from .preprocess import normalize_data
from .io_helper import save_windowed_dataset_hdf5
import numpy as np
import os
import re
import json
from tqdm import tqdm


def get_file_list(root, ext=""):
    f_list = []

    for r, d, f in os.walk(root):
        for fname in f:
            if fname.endswith(ext):
                f_list.append("{}/{}".format(r, fname))
    return f_list


def prepare_data(root, accelerometer=True, gyroscope=True, orientation=False, stroke=False, merge_clap_null=True, verbose=1):
    data = []
    labels = []
    subjects = []

    label_info = dict()
    subject_list = set()
    file_list = get_file_list(root, ext=".mat")

    pbar = tqdm(file_list, desc="Reading mat files ...")
    for i, fpath in enumerate(pbar):
        subject_name = re.search("(S[0-9]+_|[sS]troke[0-9]+)", fpath).group().replace("_", "").lower()

        if (stroke and not subject_name.startswith("stroke")) or \
                (not stroke and subject_name.startswith("stroke")):
            continue

        pbar.set_description("Reading {} ...".format(fpath))

        subject_list.add(subject_name)

        try:
            csv, video, segment = read_mat_file(fpath)

            for sensor_data, label, name in zip(segment['segment_sensor_data'], segment['segment_label'], segment['segment_name']):
                name = name.replace(" ", "").lower().replace("pck", "pick")

                # Fix wrong label START
                erratas = {
                        'clapnull': 'clap',
                        'openrefrigv': 'openrefrig',
                        '청소기밀기': 'vacuum',
                        'clapl': 'clap'
                    }

                if name in erratas.keys():
                    name = erratas[name]
                # Fix wrong label END

                if merge_clap_null:
                    name = name.replace("clap", "null")

                if name not in label_info.keys():
                    label_info[name] = len(label_info)

                data.append(sensor_data)
                subjects.append(subject_name)
                labels.append(label_info[name])

        except ValueError:
            print("Read Error! {}".format(fpath))

    subject_list = sorted(subject_list)

    if verbose > 0:
        print("Label information")
        inv_label = {v: k for k, v in label_info.items()}
        label_hist = np.histogram(labels, bins=len(label_info)-1)

        for hist, label in zip(label_hist[0], label_hist[1]):
            print("{}: {}, ".format(inv_label[label], hist), end="")
        print("")

    sensor_idx = []
    if accelerometer:
        sensor_idx += [0, 1, 2]
    if gyroscope:
        sensor_idx += [3, 4, 5]
    if orientation:
        sensor_idx += [6, 7, 8]

    for i in range(len(data)):
        data[i] = data[i][:, sensor_idx, :]

    return data, labels, subjects, label_info, subject_list


def generate_training_test_data(data, label, subjects, subject_list, training_portion=0.7, shuffle=False, cv=-1, n_cv=1, verbose=0):

    training_size = int(len(subject_list)*training_portion)

    sub_list = list(subject_list.copy())
    if shuffle:
        np.random.shuffle(sub_list)

    if cv > 0:
        test_size = int(len(subject_list) * 1/cv)

        idx1 = int(len(subject_list) * (1-(n_cv/cv)))
        idx2 = min(idx1+test_size, len(subject_list))

        training_subjects = sub_list[0:idx1] + sub_list[idx2:]
    else:
        training_subjects = sub_list[:training_size]

    if verbose > 0:
        for sub in subject_list:
            if sub in training_subjects:
                print(" {} , ".format(sub), end="")
            else:
                print("|{}|, ".format(sub), end="")
        print("")

    training_data = []
    training_label = []
    training_subject = []
    test_data = []
    test_label = []
    test_subject = []

    for i in range(len(data)):
        if subjects[i] in training_subjects:
            training_data.append(data[i])
            training_label.append(label[i])
            training_subject.append(subjects[i])
        else:
            test_data.append(data[i])
            test_label.append(label[i])
            test_subject.append(subjects[i])

    return training_data, training_label, training_subject, test_data, test_label, test_subject


#TODO : make_training_data without sliding window
def make_training_data(data_root, save_root, window_size=300, stride=1, chunk_size=50,
                       normalize_axis=True, normalize_max=1,
                       merge_clap_null=True, training_portion=0.8, shuffle=True,
                       verbose=1):
    data, label, subjects, label_info, subject_list = prepare_data(data_root, merge_clap_null=merge_clap_null)

    if normalize_axis:
        data = normalize_data(data, norm_max=normalize_max)

    training_data, training_label, training_subject, test_data, test_label, test_subject = \
        generate_training_test_data(data, label, subjects, subject_list, training_portion=training_portion,
                                    shuffle=shuffle, verbose=verbose)

    save_windowed_dataset_hdf5(training_data, training_label, test_data, test_label,
                               window_size=window_size, stride=stride, save_root=save_root,
                               chunk_size=chunk_size)

    with open("{}label_info.json".format(save_root), "w") as f:
        json.dump(label_info, f)


if __name__ == "__main__":
    pass
